-- | Contains the core Gulik evaluator.
module Gulik.Eval(-- * Interp monad actions
                  -- | These are useful for building buildins.
                  push, pop,
                  -- * Evaluation functions
                  evalEnv, eval, exec) where

-- Import a bunch of things, with care towards name collisions.
import Gulik.Types
import Prelude hiding(lookup)
import Data.Map as Map hiding(map)
import Control.Monad.State
import Control.Monad.Error

-- | Push a value onto the interpreter stack.
push :: Value     -- ^ Value to push. 
     -> Interp () -- ^ Resulting Interp monad action.
push v = do a <- get
            put (a { stateStack = v : (stateStack a) })

-- | Pop a value from the interpreter stack.
pop :: Interp Value -- ^ Interp monad action containing popped value.
pop = do a <- get
         case stateStack a of
           [] -> throwError StackUnderflow
           x:xs -> do put (a { stateStack = xs })
                      return x

-- Apply the effect of one expression to the stack machine.
step :: Expr -> Interp ()
-- Basic types are simply pushed onto the stack.
step (ExprDouble x)  = push (ValueDouble x)
step (ExprString x)  = push (ValueString x)
step (ExprInteger x) = push (ValueInteger x)
-- Names are looked up, and the corresponding value pushed onto the stack.
step (ExprName n)    = do a <- get
                          if (member n (stateEnv a)) then do v <- lookup n (stateEnv a)
                                                             push v
                                                     else throwError (UndefinedSymbol n)
-- Name binds cause a value to be popped, and the environment updated.
step (ExprBind n)    = do v <- pop
                          a <- get
                          put (a { stateEnv = insert n v (stateEnv a) })
-- Closures are pushed onto the stack.
step (ExprBody b)    = do a <- get
                          push (ValueClosure { valueEnv = (stateEnv a), valueBody = b })

{- Application causes a closure to be popped, then executed.  In the
   case of a non-builtin closure, eval is called recursively, so that
   the machine state does not have to track nested environments. -}
-- N.B.: this means that builtin closures can modify the environment,
-- although doing so breaks semantics.
step ExprApply       = do c <- pop
                          a <- get
                          case c of
                            ValueClosure { valueEnv = e, valueBody = b } -> do res <- lift (eval e (stateStack a) b)
                                                                               put (a { stateStack = res })
                            ValueBuiltin { valueBuiltin = b} -> b
                            otherwise -> throwError (InvalidApply c)

{- | Pure functional interface to the interpreter: takes an environment,
     an initial stack, and a list of expressions; returns the new stack
     and environment, or an error. -}
evalEnv :: Env                               -- ^ Initial environment.
        -> [Value]                           -- ^ Initial stack.
        -> [Expr]                            -- ^ Expressions to evaluate.
        -> Either InterpError ([Value], Env) -- ^ Error, or resulting (stack, environment) pair.
evalEnv env st pg = case result of
                      Left e -> Left e
                      Right ((), InterpState { stateStack = nst, stateEnv = nenv }) -> Right (nst, nenv)
    where
      startState = InterpState { stateStack = st, stateEnv = env }
      result     = runStateT (mapM_ step pg) startState

-- | Version of evalEnv which discards environment.
eval :: Env                        -- ^ Initial environment.
     -> [Value]                    -- ^ Initial stack.
     -> [Expr]                     -- ^ Expressions to evaluate.
     -> Either InterpError [Value] -- ^ Error, or resulting stack.
eval env st pg = (evalEnv env st pg) >>= (return . fst)

-- | Version of the above which discards stack.
exec :: Env                    -- ^ Initial environment.
     -> [Value]                -- ^ Initial stack.
     -> [Expr]                 -- ^ Expressions to evaluate.
     -> Either InterpError Env -- ^ Error, or resulting environment.
exec env st pg = (evalEnv env st pg) >>= (return . snd)
